import logging
from json import dump as j_dumps
from json import load as j_load
from os import mkdir
from os.path import join, abspath
from time import strftime

import pandas as pd
from dill import dump as d_dump
from dill import load as d_load
from numpy import vstack, pad, diff, frombuffer, round, mean, ndarray, std, histogram, exp, percentile
from sklearn.mixture import GaussianMixture
from tqdm import tqdm

import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('TkAgg')

from . import normalisation

# tools
def save_thunder(obj, path, filename='thunder.d'):
    """
    save a thunder object to a path using the dill package
    :param obj:
    :param path:
    :param filename:
    :return:
    """
    logging.debug(f'saving using dill {filename}')
    d_dump(obj, open(abspath(join(path, filename)), 'wb'))


def load_thunder(path):
    """
    load a dill dumped object
    :param path:
    :return:
    """
    logging.debug('loading using dill')
    obj = d_load(open(path, 'rb'))
    return obj


def save_plot(plot, path='.', figname='figure.svg'):
    """
    save a plot as a svg
    :param plot:
    :param path:
    :param figname:
    :return:
    """
    logging.debug(f'saving figure {figname}')
    plot.savefig(join(path, figname), transparent=True, format='svg')


def save_fit_report(obj, path, filename="report.json"):
    """
    save a fit report dictionary as a json
    :param obj:
    :param path:
    :param filename:
    :return:
    """
    logging.debug(f'saving report {filename}')
    j_dumps(obj, open(abspath(join(path, filename)), 'w'), indent=4)


def find_closest_indices(list1, list2):
    """
    given two lists returns list of indices indicating which indices in list 1 match each index in list 2. i.e. index 0
    of the returned list indicates which index in list 1 matches index 0 in list 2
    :param list1: a list of numbers
    :param list2: a list of numbers
    :return: a list of indices of matches
    """
    try:
        list_of_matching_indices = [min(range(len(list1)), key=lambda i: abs(
            list1[i] - cent)) for cent in list2]  # the lambda function is what min uses
    except ValueError:
        print('this dataset has no values!')
        return
    return list_of_matching_indices


def normalise_all(y_bg_rem, bg, y_raw):
    """
    given data with background removed, a background and the raw data, normalise the data with the bg removed and then
    normalise the others according to this normalisation (i.e. use the mean and std from that)
    :param y_bg_rem: np array of data after bg removed
    :param bg: np array of bg
    :param y_raw: np array of data before bg removed
    :return: data np arrays normalised by svn normalisaiton
    """
    logging.debug('normalising many objects')
    y_data_bg_rm, (mean_y_data, std_dev) = normalisation.svn(
        y_bg_rem)  # normalise the data
    # normalise with data from bg subtracted data
    background, _ = normalisation.svn(bg, mean_y_data, std_dev)
    # normalise with data from bg subtracted data
    y_data_norm, _ = normalisation.svn(y_raw, mean_y_data, std_dev)

    return y_data_bg_rm, background, y_data_norm


def safe_list_get(l, idx, default):
    """fetch items safely from a list, if it isn't long enough return a default value"""
    try:
        return l[idx]
    except IndexError:
        return default


def sharpening_routine(x_data, y_data):
    """
    a function to run a user guided routine to sharpen data using peak_sharpening function
    :param x_data: np array of x data
    :param y_data: np array of y data
    :return: y_data sharpened by user chosen factors
    """
    sharpening_factor = (0, 0)
    res_enhanced = y_data
    while True:
        plt.plot(x_data, res_enhanced)
        print(
            f"Do you want to sharpen the peaks to help find components? Note this will not edit the actual data. "
            f"Current sharpening factor is: {sharpening_factor}")
        plt.show()
        ans = input(
            "Please enter the method (either 'power' or 'deriv'), then a comma, then a new sharpening factors "
            "(comma seperated if mutliple i.e. for derivative), "
            "or type y to continue with the current factor")
        if ans == 'y':
            plt.close()
            return y_data
        else:
            try:
                ans = ans.split(',')
                _type = ans[0]
                ans = ans[1:]
                sharpening_factor = [float(fac) for fac in ans]
                res_enhanced = peak_sharpening(
                    y_data, _type, sharpening_factor)
            except BaseException:
                print("You entered an incorrect answer! Trying again...")


def peak_sharpening(y_data, _type, sharpening_factor):
    """
    function to sharpen peaks. can use power or derivative methods
    :param y_data: np array of y data
    :param _type: the type of sharpening to be used
    :param sharpening_factor: the factors to sharpen with
    :return: the data sharpened
    """
    if _type == 'power':
        res_enhanced = y_data ** sharpening_factor[0]  # raise to the power
    elif _type == 'deriv':
        y_double_prime = pad(diff(y_data, n=2), (0, 2), 'constant')
        y_4_prime = pad(diff(y_data, n=4), (0, 4), 'constant')
        res_enhanced = y_data - sharpening_factor[0] * y_double_prime + sharpening_factor[
            1] * y_4_prime  # this is the original data minus its
        # derivative multiplied by some factor
    else:
        raise ValueError("enter a correct type")
    return res_enhanced

# tools

# user inputs and loading etc


def load_data(datapath, x_ind, y_ind, e_ind=None):
    """
    load in data as an np arra. can load from a csv or hs5 pandas
    :param datapath: where to get data from
    :param x_ind: which column to find x data
    :param y_ind: which column to find y data
    :param e_ind: optional, which column to find error data
    :return: np arrays of the data loaded with nan values dropped across all values (note might break with error values)
    """
    logging.debug('loading data')
    if '.h5' in datapath:  # if the data is already stored as a pandas df
        store = pd.HDFStore(datapath)
        keys = store.keys()
        if len(keys) > 1:
            logging.warning(
                "Too many keys in the hdfstore, will assume all should be concated")
            logging.warning("not sure this concat works yet")
            # not sure this will work! concat all keys dfs together
            data = store.concat([store[key] for key in keys])
        else:
            # if only one key then we use it as the datafile
            data = store[keys[0]]
    else:  # its a txt or csv file
        # load in, works for .txt and .csv
        data = pd.read_csv(datapath, header=None, sep='\t', dtype='float')
        if len(data.columns) < 2:
            # load in, works for .txt and .csv
            data = pd.read_csv(
                datapath,
                header=None,
                sep=r'\s+',
                dtype='float')
        # this needs to be made more flexible/user defined
    if e_ind:  # if we have specified this column then we use it, otherwise just x and y
        assert (len(data.columns) >=
                2), "You have specified an e_ind but there are less than 3 columns in the data"
        e_data = data[e_ind].values
    else:
        e_data = None

    data.dropna()  # drop any rows with NaN etc in them

    x_data = data[x_ind].values
    y_data = data[y_ind].values

    return x_data, y_data, e_data


def map_unique_coords(x_data, y_data, x_coords, y_coords):
    """
    function to get the unique coordinate values out from np arrays of data and coordinates
    :param x_data:np array of x data
    :param y_data: np array of y data
    :param x_coords: np array of x coordinates, index corresponds to x and y data
    :param y_coords: np array of y coordinates, index corresponds to x and y data
    :return: lists of np arrays, each element matches, x_coords and y_coords contain lists of numbers only. so e.g.
    index 0 of all has the x and y coordinates in coords lists at 0, and x and y data in those lists at 0 as np arrays
    """
    logging.debug('parsing coordinates')
    data = vstack((x_coords, y_coords, x_data, y_data)
                  ).transpose()  # now have columns as the data
    df = pd.DataFrame(
        data=data,
        columns=[
            'x_coords',
            'y_coords',
            'x_data',
            'y_data'])
    # get a dictionary of the unique values for
    unique_dict = dict(tuple(df.groupby(['x_coords', 'y_coords'])))
    # coordinates (as tuples of (x,y)) and then the whole df rows for these
    # values

    x_data, y_data, x_coords, y_coords = [], [], [], []
    for key in unique_dict.keys():
        x_data_ = unique_dict[key]['x_data'].values  # get the x_data
        x_data.append(x_data_)
        y_data_ = unique_dict[key]['y_data'].values
        y_data.append(y_data_)
        x_coords.append(key[0])
        y_coords.append(key[1])

    return x_data, y_data, x_coords, y_coords


def parse_param_file(filepath='./params.txt'):
    """
    parse a params file which we assume is a dictionary
    :param filepath: str: path to params file
    :return: dictionary of paramters
    """
    # maybe use json loads if you end up writing parameter files non-manually
    logging.debug('parsing params file')
    with open(filepath, 'r') as f:
        arguments = j_load(f)
        f.close()

    # TODO: add some checks to user passed data
    return arguments


def parse_args(arg):
    """
    convert argparse arguments into a dictionary for consistency later
    :param arg: argparse parsed args
    :return: dictionary of parameters
    """
    logging.debug('parsing args')
    arguments = {}
    arguments['x_ind'] = arg.x_ind
    arguments['y_ind'] = arg.y_ind
    arguments['e_ind'] = arg.e_ind
    arguments['datapath'] = arg.datapath
    arguments['no_peaks'] = arg.no_peaks
    arguments['background'] = arg.background
    arguments['scarf_params'] = arg.scarf_params
    arguments['peak_types'] = arg.peak_types
    arguments['peak_centres'] = arg.peak_centres
    arguments['peak_widths'] = arg.peak_widths
    arguments['peak_amps'] = arg.peak_amps
    arguments['tightness'] = arg.tightness
    arguments['bounds'] = arg.bounds

    # TODO: add some checks to user passed data
    return arguments


def make_dir(dirname, i=1):
    """
    function to make a directory, recursively adding _new if that name already exists
    :param dirname: str: name of directory to create
    :param i: the run number we are on
    :return: str: the directory name which was available, and all subsequent data should be saved in
    """
    logging.debug('making dir')
    try:
        mkdir(f'{dirname}')
    except FileExistsError as e:
        dirname = make_dir(f'{dirname}_new', i + 1)
        if i == 1:
            print(e, f'. So I named the file: {dirname}')
        return dirname
    return dirname


def clip_data(x_data, y_data, clips=None):
    """
    given data either clip it or run a user guided routine to clip it
    :param x_data: np array of x data
    :param y_data: np array of y data
    :param clips: either none or a list of two values which are left and right clips in terms of x values
    :return: the indices of the clips to use
    """
    logging.debug('clipping data')
    if clips:
        clip_left, clip_right = clips
        clip_left = find_closest_indices(list(x_data), [clip_left])[0]
        clip_right = find_closest_indices(list(x_data), [clip_right])[0]
    else:
        clip_left, clip_right = 0, len(x_data) - 1
        while True:
            fig, ax = plt.subplots()
            ax.plot(x_data[clip_left:clip_right], y_data[clip_left:clip_right])
            print(
                f"Removing background, please type two x values seperated by a space for the clips. \n"
                f"Current values are: {x_data[clip_left]}, {x_data[clip_right]}. \n"
                f"PLEASE MAKE SURE YOU ENTER IN THE SAME ORDER AS HERE. i.e. if first value is larger than right then the "
                f"first value will be the large x_clip second small")
            plt.show(block=True)
            ans = input(
                "If you are happy with the clips type y. If not then please type a new pair of values ")
            if ans == 'y':
                break
            else:
                try:
                    ans = ans.split(' ')
                    if len(ans) != 2:
                        raise ValueError(
                            "The tuple was more than two elements long")
                    clip_left = float(ans[0])
                    clip_left = find_closest_indices(
                        list(x_data), [clip_left])[0]
                    clip_right = float(ans[1])
                    clip_right = find_closest_indices(
                        list(x_data), [clip_right])[0]
                except ValueError:
                    print("You entered an incorrect answer! Trying again...")

        plt.close()
    return clip_left, clip_right


def apply_func(key_kwargs_, func):
    """
    given some keywords and a function call the function
    :param key_kwargs_: a tuple of (key, args) to use
    :param func: function to call
    :return: the key for this and the value returned from calling the func
    """
    key = key_kwargs_[0]
    kwargs_ = key_kwargs_[1]
    val = func(*kwargs_)
    return key, val


def setup_logger(log_name):
    """
    function to setup a logger to save to file
    :param log_name:
    :return:
    """
    curr_time = strftime('%d_%m_%Y__%H;%M')
    log_filename = f"{log_name}_{curr_time}.log"
    logging.getLogger().setLevel(logging.DEBUG)
    logger = logging.getLogger('')
    logger.handlers = []
    logging.basicConfig(filename=log_filename, level=logging.DEBUG)
    logging.info('have read in user arguments')
    return log_filename


def gif_maker(bag, filename):
    """
    function to make a gif of all the plots (with no uncertainty) and save it at filename
    :param bag: a dictionary of the thunder objects which have been fit etc and are to be plotted
    :param filename: where to save the gif
    :return:
    """
    bags = iter(bag.values())

    def update(i):
        thund = next(bags)
        ax, fig = thund.plot_all(plot_unc=False)
        plt.text(
            0.1,
            0.9,
            f'PLOT_{i}',
            horizontalalignment='center',
            verticalalignment='center',
            transform=ax.transAxes)
        fig.canvas.draw()
        img = frombuffer(fig.canvas.tostring_rgb(), dtype='uint8')
        img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        plt.close()
        return img

    import imageio
    print("Creating gif")
    imageio.mimsave(filename, tqdm([update(i) for i in range(len(bag))]), fps=2)
#

def cosmic_rays(y_data, width_threshold=3):
    """
    this currently fails if data is sloping and cosmic ray is below max values. change to sloping mad in future
    :param y_data: np array of data
    :return:
    """
    """
    import numpy as np
    from scipy.signal import find_peaks
    import ipdb
    ipdb.set_trace()
    peaks, properties = find_peaks(y_data, width=0, rel_height=0.8) # find peaks with a minimum prominence of zero -
    # finds widths at 0.8 the height of peaks - quick experiment shows this is optimal for a random example
    # - needs more testing
    prominences = properties['prominences']
    peak_ind = np.argwhere(prominences > np.percentile(prominences, 99))[:,0] # indices of peaks with prominence in 99th percentile
    y_data_peak_indices = np.take(peaks, peak_ind)

    widths = np.take(properties['widths'], peak_ind) # these are the widths with this peak prominence
    np.argwhere(widths < width_threshold) # this may fail often?

    """



    """
    mad = np.median(abs(y_data - np.median(y_data))) # median abs deviation of the data
    y_mad = abs(y_data - np.median(y_data)) / mad # abs mdeviation from median divided by mad to get ratio
    cutoff = np.percentile(y_mad, 99.5) # what values make up the 99th percentile, above this are rare!
    bad_indices = np.argwhere(y_mad > cutoff)
    for i in bad_indices[:,0]:
        y_data[i] = np.mean(y_data[i-10:i+10]) # set as the mean value in a window around the peak
    """
    #### ideas:
    ### detection:
    ## r-pca as from stanford - probably not good here. https://medium.com/netflix-techblog/rad-outlier-detection-on-big-data-d6b0494371cc : https://github.com/dganguli/robust-pca/blob/master/r_pca.py
    ## differentiate in time across multiple spectra, use that to detect spikes - for single spectra do user guided 'zap' function?
    ## just smoothing? or smoothing and then a residual threshold?
    ## find peaks using scipy and delete low width peaks, doesn't work if small though
    ## https://journals.sagepub.com/doi/pdf/10.1366/000370207781745847 - simulate the data and set a threshold for
    # residual, replace pixels with simulated. this would involve a cosmic spike removal step at the end of fitting.
    # if fitting has already failed then wouldn't work. could be a bit delicate
    ## asto guy: https://cosmicrayapp.com/2017/02/03/signal-processing-details/
    ## mad - also captures peaks in general so no bueno
    ## https://pureportal.strath.ac.uk/files-asset/38783720/Littlejohn_Automated_cosmic_spike_filter.pdf
    ## https://www.osti.gov/pages/servlets/purl/1334720
    ## https://www.researchgate.net/publication/233403614_Automated_Cosmic_Spike_Filter_Optimized_for_Process_Raman_Spectroscopy


    return y_data


def histogram_func(u_vals:ndarray, x_label, gmm=False, f=None, ax=None, bins='auto'):
    assert(len(u_vals.shape) == 1), "The nd array passed to histogram func must be a 1d ndarray"
    if not ax or f: # then create the figure
        f = plt.figure()
        ax = f.add_subplot(111)
    # get the hist
    range_l, range_h = percentile(u_vals,1), percentile(u_vals,99)
    if (range_h - range_l) > 5 * (percentile(u_vals,90) - percentile(u_vals,10)):
        range_l = percentile(u_vals, 5)
        range_h = percentile(u_vals, 95)
    heights, edges = histogram(a=u_vals, bins=bins, range=(range_l, range_h ))
    widths = [0.8*(edges[i+1] - edges[i]) for i in range(len(edges) - 1)]
    edges = edges[:-1]
    # plot it
    ax.bar(edges, heights, width=widths, color='r', align='edge')
    ax.grid(axis='y', alpha=0.75)
    ax.set(xlabel=x_label, ylabel='Frequency')
    mu = round(mean(u_vals), 1)
    mu_unc = round(std(u_vals) / len(u_vals), 1) # standard error
    sig = round(std(u_vals), 1) # sigma

    plt.text(0.3, 0.9, r'$\mu=$' + f'{mu}' + f'pm {mu_unc}' + ', ' + r'$\sigma=$' + f'{sig}' ,
             horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
    if gmm: # then add gaussian mixture
        vals, bins = histogram(u_vals, bins='auto')
        gmm = GaussianMixture(n_components=3)
        gmm = gmm.fit(u_vals[:, None])
        plt.plot(bins[:None], exp(gmm.score_samples(bins[:, None])))
        plt.text(0.3, 0.8, r'GaussianMixture Components:', horizontalalignment='center',
                 verticalalignment='center', transform=ax.transAxes)
    return f, ax, bins



import logging

from lmfit import models
from numpy import nan

from . import utilities as utils


def get_tols(method, tol):
    """
    quick function to set the tolerances as a dictionary depending on the method being used
    :param method: str: method to use
    :param tol: tolerance as a float
    :return:
    """
    if method in ['leastsq', 'least_squares']:
        # warning: need to experiment with these/allow control maybe?
        tols = {'ftol': tol, 'xtol': tol}
    else:
        tols = {'ftol': tol}
    return tols


def prep_algo(method, tols):
    """
    depending on the algorithm either set the tolerances, give a warning or set a parameter which is depreciated to True.
    will be overhauled in future
    :param method: str: method to use
    :param tols: float tolerance
    :return:
    """
    if method == 'basinhopping' or method == 'ampgo':
        print('Warning: This is a very slow but thorough algorithm')
    if method == 'differential_evolution':
        all_bounds = True
    if method in [
        'leastsq',
        'least_squares',
        'nelder',
            'cobyla']:  # do this for all of them??
        tols = get_tols(method, tols)
    return tols


def fit_peaks(
        x_data,
        y_data,
        peak_info_dict,
        bounds,
        method='leastsq',
        tol=0.0000001):
    """
    function which does the peak fitting.
    :param x_data: np array of x data
    :param y_data: np array of y data
    :param peak_info_dict: dictionary containing information about the peaks which will be used
    :param bounds: bounds with same keys as peak_info_dict and each is a list of tuples containing lower and upper bounds
    :param method: str: minimisation method to use
    :param tol: tolerance on minimisation.
    :return: specifications dictionary for the model used, the model object iteself, the parameters of the best fitted
    values and the model_result object
    """
    impl_methods = [
        'leastsq',
        'least_squares',
        'nelder',
        'lbfgsb',
        'powell',
        'cg',
        'cobyla',
        'bfgsb',
        'differential_evolution',
        'basinhopping',
        'ampgo']
    if method not in impl_methods:
        raise ValueError(
            f"The method supplied is not supported. Available methods: {impl_methods}")
    logging.debug(f'fitting peaks:  {peak_info_dict}')

    tols = prep_algo(method, tol)  # prep for the fitting
    # build a correctly formatted dictionary for fitting
    model_specs = build_specs(peak_info_dict, bounds)
    # generate the composite model for fitting
    model, peak_params = generate_model(model_specs)

    if method in ['leastsq', 'least_squares', 'nelder', 'cobyla']:
        peaks = model.fit(
            y_data,
            peak_params,
            x=x_data,
            method=method,
            fit_kws=tols)  # fit the model to the data
        if not peaks.success:  # then raise tolerance and try again
            print(
                'peaks failed to fit, raising tolerance by one order magnitude and trying again')
            # try raising the tolerance if it fails by one order
            tols = {key: value * 10 for key, value in tols.items()}
            peaks = model.fit(
                y_data,
                peak_params,
                x=x_data,
                method=method,
                fit_kws=tols)
    else:
        peaks = model.fit(y_data, peak_params, x=x_data, method=method)

    # what are the best parameters used # shoudl create a function as an
    # abstraction barrier for this
    peak_params = peaks.best_values
    if not peaks.success:  # then didn't fit for some reason, so set everything to nan in results
        print('peaks failed to fit')
        peak_params = {key: nan for key in peak_params}

    # really should clean this up but allows flexibility for user by passing
    # the actual objects back too
    return model_specs, model, peak_params, peaks


def safe_list_get(l, idx, default):
    """fetch items safely from a list, if it isn't long enough return a default value"""
    try:
        return l[idx]
    except IndexError:
        return default

# Call this with correct arguments, should always call with all the arguments!


def build_specs(peak_info_dict: dict = {}, bounds: dict = {}):
    """
    should be depreciated in future. can get by with bypassing this, currently acts to cleanse inputs so maybe could be
    shortened/changed with that goal in mind
    :param peak_info_dict:
    :param bounds:
    :return:
    """

    """Build a specs list, which has the peak specification details for each element corresponding to a peak to fit.
    Note that each element is a dictionary containing either None, or a value for that peak i.e. the parameter guess,
    or the bounds for that value. Only some will be valid for each peak type, which will be handled elsewhere"""
    logging.debug('building specs')
    type_ = peak_info_dict.get('type', ())
    if 'SplitLorentzianModel' in type_ or 'ExponentialGaussianModel' in type_:
        raise NotImplementedError(
            "No support for this model yet: polynomialmodel, ExponentialGaussianModel and splitlorentzianmodel")
    if 'ExpressionModel' in type_:
        raise NotImplementedError("no support for expression models yet!")
    center = peak_info_dict.get('center', ())
    amplitude = peak_info_dict.get('amplitude', ())
    sigma = peak_info_dict.get('sigma', ())
    sigma_l = peak_info_dict.get('sigma_l', ())
    sigma_r = peak_info_dict.get('sigma_r', ())
    gamma = peak_info_dict.get('gamma', ())
    fraction = peak_info_dict.get('fraction', ())
    beta = peak_info_dict.get('beta', ())
    exponent = peak_info_dict.get('exponent', ())
    q = peak_info_dict.get('q', ())
    c = peak_info_dict.get('c', ())
    intercept = peak_info_dict.get('intercept', ())
    slope = peak_info_dict.get('slope', ())
    a = peak_info_dict.get('a', ())
    b = peak_info_dict.get('b', ())
    degree = peak_info_dict.get('degree', ())
    center1 = peak_info_dict.get('center1', ())
    center2 = peak_info_dict.get('center2', ())
    sigma1 = peak_info_dict.get('sigma1', ())
    sigma2 = peak_info_dict.get('sigma2', ())
    decay = peak_info_dict.get('decay', ())
    expr = peak_info_dict.get('expr', ())
    form = peak_info_dict.get('form', ())
    c0 = peak_info_dict.get('c0', ())
    c1 = peak_info_dict.get('c1', ())
    c2 = peak_info_dict.get('c2', ())
    c3 = peak_info_dict.get('c3', ())
    c4 = peak_info_dict.get('c4', ())
    c5 = peak_info_dict.get('c5', ())
    c6 = peak_info_dict.get('c6', ())
    c7 = peak_info_dict.get('c7', ())

    specs = [
        {
            'type': type_[i],
            'params': {'center': utils.safe_list_get(center, i, None),
                       'amplitude': utils.safe_list_get(amplitude, i, None),
                       'sigma': utils.safe_list_get(sigma, i, None),
                       'sigma_r': utils.safe_list_get(sigma_r, i, None),
                       'sigma_l': utils.safe_list_get(sigma_l, i, None),
                       'gamma': utils.safe_list_get(gamma, i, None),
                       'fraction': utils.safe_list_get(fraction, i, None),
                       'beta': utils.safe_list_get(beta, i, None),
                       'exponent': utils.safe_list_get(exponent, i, None),
                       'q': utils.safe_list_get(q, i, None),
                       'c': utils.safe_list_get(c, i, None),
                       'intercept': utils.safe_list_get(intercept, i, None),
                       'slope': utils.safe_list_get(slope, i, None),
                       'a': utils.safe_list_get(a, i, None),
                       'b': utils.safe_list_get(b, i, None),
                       'center1': utils.safe_list_get(center1, i, None),
                       'center2': utils.safe_list_get(center2, i, None),
                       'sigma1': utils.safe_list_get(sigma1, i, None),
                       'sigma2': utils.safe_list_get(sigma2, i, None),
                       'decay': utils.safe_list_get(decay, i, None),
                       'c0': utils.safe_list_get(c0, i, None),
                       'c1': utils.safe_list_get(c1, i, None),
                       'c2': utils.safe_list_get(c2, i, None),
                       'c3': utils.safe_list_get(c3, i, None),
                       'c4': utils.safe_list_get(c4, i, None),
                       'c5': utils.safe_list_get(c5, i, None),
                       'c6': utils.safe_list_get(c6, i, None),
                       'c7': utils.safe_list_get(c7, i, None)
                       },
            'bounds': {'center': utils.safe_list_get(bounds.get('center', []), i, (None, None)),
                       'amplitude': utils.safe_list_get(bounds.get('amplitude', []), i, (None, None)),
                       'sigma': utils.safe_list_get(bounds.get('sigma', []), i, (None, None)),
                       'sigma_r': utils.safe_list_get(bounds.get('sigma_r', []), i, (None, None)),
                       'gamma': utils.safe_list_get(bounds.get('gamma', []), i, (None, None)),
                       'fraction': utils.safe_list_get(bounds.get('fraction', []), i, (None, None)),
                       'beta': utils.safe_list_get(bounds.get('beta', []), i, (None, None)),
                       'exponent': utils.safe_list_get(bounds.get('exponent', []), i, (None, None)),
                       'q': utils.safe_list_get(bounds.get('q', []), i, (None, None)),
                       'c': utils.safe_list_get(bounds.get('c', []), i, (None, None)),
                       'intercept': utils.safe_list_get(bounds.get('intercept', []), i, (None, None)),
                       'slope': utils.safe_list_get(bounds.get('slope', []), i, (None, None)),
                       'a': utils.safe_list_get(bounds.get('a', []), i, (None, None)),
                       'b': utils.safe_list_get(bounds.get('b', []), i, (None, None)),
                       'center1': utils.safe_list_get(bounds.get('center1', []), i, (None, None)),
                       'center2': utils.safe_list_get(bounds.get('center2', []), i, (None, None)),
                       'sigma1': utils.safe_list_get(bounds.get('sigma1', []), i, (None, None)),
                       'sigma2': utils.safe_list_get(bounds.get('sigma2', []), i, (None, None)),
                       'decay': utils.safe_list_get(bounds.get('decay', []), i, (None, None)),
                       'c0': utils.safe_list_get(bounds.get('c0', []), i, (None, None)),
                       'c1': utils.safe_list_get(bounds.get('c0', []), i, (None, None)),
                       'c2': utils.safe_list_get(bounds.get('c0', []), i, (None, None)),
                       'c3': utils.safe_list_get(bounds.get('c0', []), i, (None, None)),
                       'c4': utils.safe_list_get(bounds.get('c0', []), i, (None, None)),
                       'c5': utils.safe_list_get(bounds.get('c0', []), i, (None, None)),
                       'c6': utils.safe_list_get(bounds.get('c0', []), i, (None, None)),
                       'c7': utils.safe_list_get(bounds.get('c0', []), i, (None, None))
                       },
            'expr': utils.safe_list_get(expr, i, None),
            'form': utils.safe_list_get(form, i, None),
            'degree': utils.safe_list_get(degree, i, None)
        }
        for i in range(len(type_))
    ]

    return specs


def generate_model(model_specs):
    """
    generate a composite model given information of the models to create, their guesses of params and bounds on parameters
    :param model_specs: a dictionary containing all the info on the model
    :return: a composite lmfit model
    """
    logging.debug('generating model specs')
    composite_model = None
    params = None
    for i, spec in enumerate(model_specs):
        prefix = f'm{i}__'
        if spec['type'] == 'ExpressionModel':
            expr = spec['expr']
            model = models.ExpressionModel(expr)
        elif spec['type'] in ['StepModel', 'RectangleModel']:
            form = spec['form']
            model = getattr(models, spec['type'])(prefix=prefix, form=form)
        elif spec['type'] == 'PolynomialModel':
            model = getattr(
                models,
                spec['type'])(
                prefix=prefix,
                degree=spec['degree'])
        else:
            # generate the lmfit model based on the type specified
            model = getattr(models, spec['type'])(prefix=prefix)
        # call another function to decide what to do
        model = decide_model_actions(spec, model)
        model_params = model.make_params()  # make the params object
        if params is None:  # first loop
            params = model_params
            composite_model = model
        else:  # subsequent loops
            params.update(model_params)
            composite_model = composite_model + model
    return composite_model, params


def decide_model_actions(spec, model):
    """
    either set the param hints or the bounds depending on the spec dict
    :param spec: dictionary of the params to be set on the model
    :param model: lmfit model object
    :return: the updated model object
    """
    for param_key, param_value in spec['params'].items():
        if param_value:  # then set this value
            model.set_param_hint(param_key, value=param_value)
    for bound_key, bound_value in spec['bounds'].items():
        if bound_value[0]:  # then set lower bound
            model.set_param_hint(bound_key, min=bound_value[0])
        if bound_value[1]:  # then set upper bound
            model.set_param_hint(bound_key, max=bound_value[1])
    return model



from setuptools import setup, find_packages

def readme():
    with open("README.rst") as f:
        return f.read()

setup(
    name='thunderfit',
    python_requires='>3.6',
    version='1.7.2.3',
    description='Thunderfit fitting code',
    long_description=readme(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    keywords='raman analysis fitting',
    url='https://github.com/Mitchwatts93/thunderfit',
    author='https://github.com/Mitchwatts93',
    author_email='',
    license='MIT',
    packages=find_packages(),
    install_requires = ['jsonschema==2.6.0',
        'dill==0.2.9',
        'scipy==1.1.0',
        'numpy==1.15.1',
        'matplotlib==2.2.3',
        'pandas==0.23.4',
        'lmfit==0.9.13',
        'tqdm==4.26.0',
        'asteval==0.9.13',
        'typing==3.6.6',
        'imageio ==2.5.0'
        ],
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['nose'],
    entry_points={
        "console_scripts": ['ramananalyse = thunderfit.raman_routine:main', 'mapscan = thunderfit.map_scan:main',
                            'genparams = thunderfit.create_params:main']},
    )



from __future__ import division, print_function

import numpy as np

# https://github.com/dganguli/robust-pca/blob/master/r_pca.py

try:
    from pylab import plt
except ImportError:
    print('Unable to import pylab. R_pca.plot_fit() will not work.')

try:
    # Python 2: 'xrange' is the iterative version
    range = xrange
except NameError:
    # Python 3: 'range' is iterative - no need for 'xrange'
    pass


class R_pca:

    def __init__(self, D, mu=None, lmbda=None):
        self.D = D
        self.S = np.zeros(self.D.shape)
        self.Y = np.zeros(self.D.shape)

        if mu:
            self.mu = mu
        else:
            self.mu = np.prod(self.D.shape) / (4 * self.frobenius_norm(self.D))

        self.mu_inv = 1 / self.mu

        if lmbda:
            self.lmbda = lmbda
        else:
            self.lmbda = 1 / np.sqrt(np.max(self.D.shape))

    @staticmethod
    def frobenius_norm(M):
        return np.linalg.norm(M, ord='fro')

    @staticmethod
    def shrink(M, tau):
        return np.sign(M) * np.maximum((np.abs(M) - tau), np.zeros(M.shape))

    def svd_threshold(self, M, tau):
        U, S, V = np.linalg.svd(M, full_matrices=False)
        return np.dot(U, np.dot(np.diag(self.shrink(S, tau)), V))

    def fit(self, tol=None, max_iter=1000, iter_print=100):
        iter = 0
        err = np.Inf
        Sk = self.S
        Yk = self.Y
        Lk = np.zeros(self.D.shape)

        if tol:
            _tol = tol
        else:
            _tol = 1E-7 * self.frobenius_norm(self.D)

        while (err > _tol) and iter < max_iter:
            Lk = self.svd_threshold(
                self.D - Sk + self.mu_inv * Yk, self.mu_inv)
            Sk = self.shrink(
                self.D - Lk + (self.mu_inv * Yk), self.mu_inv * self.lmbda)
            Yk = Yk + self.mu * (self.D - Lk - Sk)
            err = self.frobenius_norm(self.D - Lk - Sk)
            iter += 1
            if (iter % iter_print) == 0 or iter == 1 or iter > max_iter or err <= _tol:
                print('iteration: {0}, error: {1}'.format(iter, err))

        self.L = Lk
        self.S = Sk
        return Lk, Sk

    def plot_fit(self, size=None, tol=0.1, axis_on=True):

        n, d = self.D.shape

        if size:
            nrows, ncols = size
        else:
            sq = np.ceil(np.sqrt(n))
            nrows = int(sq)
            ncols = int(sq)

        ymin = np.nanmin(self.D)
        ymax = np.nanmax(self.D)
        print('ymin: {0}, ymax: {1}'.format(ymin, ymax))

        numplots = np.min([n, nrows * ncols])
        plt.figure()

        for n in range(numplots):
            plt.subplot(nrows, ncols, n + 1)
            plt.ylim((ymin - tol, ymax + tol))
            plt.plot(self.L[n, :] + self.S[n, :], 'r')
            plt.plot(self.L[n, :], 'b')
            if not axis_on:
                plt.axis('off')